{% if fluff -%}/**********************************************************************************************
 * Filename:       {{service.name}}.c
 *
 * Description:    This file contains the implementation of the service.
 *
 *                 Generated by:
 *                 BDS version: {{pluginInfo.AppVersion}}
 *                 Plugin:      {{pluginInfo.Name}} {{ pluginInfo.Version}}
 *                 Time:        {{pluginInfo.timestamp}}
 *

 * Copyright (c) 2015, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *************************************************************************************************/


/*********************************************************************
 * INCLUDES
 */
#include <string.h>

#include "bcomdef.h"
#include "OSAL.h"
#include "linkdb.h"
#include "att.h"
#include "gatt.h"
#include "gatt_uuid.h"
#include "gattservapp.h"
#include "gapbondmgr.h"

#include "{{service.name}}.h"

/*********************************************************************
 * MACROS
 */

/*********************************************************************
 * CONSTANTS
 */

/*********************************************************************
 * TYPEDEFS
 */

/*********************************************************************
* GLOBAL VARIABLES
*/

{% endif -%}// {{service.name}} Service UUID
CONST uint8_t {{service.nameCap}}UUID[{{ service.uuid_size }}] =
{
{% if service.uuid_size == 'ATT_UUID_SIZE' -%}
  {{ service.name | upcase }}_SERV_UUID_BASE128({{ service.name | upcase }}_SERV_UUID)
{% else -%}
  LO_UINT16({{ service.name | upcase }}_SERV_UUID), HI_UINT16({{ service.name | upcase }}_SERV_UUID)
{% endif -%}
};

{% for char in service.chars -%}
// {{char.name}} UUID
CONST uint8_t {{ service.nameAbbr}}_{{ char.nameCap }}UUID[{{ char.uuid_size }}] =
{
{% if char.uuid_size == 'ATT_UUID_SIZE' -%}
  {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_UUID_BASE128({{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_UUID)
{% else -%}
  LO_UINT16({{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_UUID), HI_UINT16({{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_UUID)
{% endif -%}
};
{% endfor -%}

{% if fluff -%}/*********************************************************************
 * LOCAL VARIABLES
 */

{% endif -%}static {{service.nameCap}}CBs_t *pAppCBs = NULL;

{% if fluff -%}/*********************************************************************
* Profile Attributes - variables
*/

{% endif -%}// Service declaration
static CONST gattAttrType_t {{service.nameCap}}Decl = { {{ service.uuid_size }}, {{service.nameCap}}UUID };
{% for char in service.chars -%}

// Characteristic "{{ char.nameNice }}" Properties (for declaration)
static uint8_t {{ service.nameAbbr}}_{{ char.nameCap }}Props = {{ char.orProps }};

// Characteristic "{{ char.nameNice }}" Value variable
static uint8_t {{ service.nameAbbr}}_{{ char.nameCap }}Val[{{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_LEN] = {0};

// Length of data in characteristic "{{ char.nameNice }}" Value variable, initialized to minimal size.
static uint16_t {{ service.nameAbbr}}_{{ char.nameCap }}ValLen = {{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_LEN_MIN;
{% if char.hasCCCD -%}

// Characteristic "{{ char.nameNice }}" CCCD
static gattCharCfg_t *{{ service.nameAbbr}}_{{ char.nameCap }}Config;
{% endif -%}
{% endfor -%}

{% if fluff -%}/*********************************************************************
* Profile Attributes - Table
*/

{% endif -%}static gattAttribute_t {{service.name}}AttrTbl[] = 
{
  // {{service.name}} Service Declaration
  {
    { ATT_BT_UUID_SIZE, primaryServiceUUID },
    GATT_PERMIT_READ,
    0,
    (uint8_t *)&{{service.nameCap}}Decl
  },
{% for char in service.chars -%}
    // {{ char.nameNice }} Characteristic Declaration
    {
      { ATT_BT_UUID_SIZE, characterUUID },
      GATT_PERMIT_READ,
      0,
      &{{ service.nameAbbr}}_{{ char.nameCap }}Props
    },
      // {{ char.nameNice }} Characteristic Value
      {
        { {{ char.uuid_size }}, {{ service.nameAbbr}}_{{ char.nameCap }}UUID },
        {{ char.orPerms }},
        0,
        {{ service.nameAbbr}}_{{ char.nameCap }}Val
      },
{% if char.hasCCCD -%}
      // {{ char.nameNice }} CCCD
      {
        { ATT_BT_UUID_SIZE, clientCharCfgUUID },
        GATT_PERMIT_READ | GATT_PERMIT_WRITE,
        0,
        (uint8_t *)&{{ service.nameAbbr}}_{{ char.nameCap }}Config
      },
{% endif -%}
{% endfor -%}
};

{% if fluff -%}/*********************************************************************
 * LOCAL FUNCTIONS
 */
{% endif -%}static bStatus_t {{service.name}}_ReadAttrCB( uint16_t connHandle, gattAttribute_t *pAttr, 
                                           uint8_t *pValue, uint16_t *pLen, uint16_t offset,
                                           uint16_t maxLen, uint8_t method );
static bStatus_t {{service.name}}_WriteAttrCB( uint16_t connHandle, gattAttribute_t *pAttr,
                                            uint8_t *pValue, uint16_t len, uint16_t offset,
                                            uint8_t method );

{% if fluff -%}/*********************************************************************
 * PROFILE CALLBACKS
 */
{% endif -%}// Simple Profile Service Callbacks
CONST gattServiceCBs_t {{service.name}}CBs =
{
  {{service.name}}_ReadAttrCB,  // Read callback function pointer
  {{service.name}}_WriteAttrCB, // Write callback function pointer
  NULL                       // Authorization callback function pointer
};

{% if fluff -%}/*********************************************************************
* PUBLIC FUNCTIONS
*/

/*
 * {{ service.nameCap }}_AddService- Initializes the {{ service.nameCap }} service by registering
 *          GATT attributes with the GATT server.
 *
 */
{% endif -%}bStatus_t {{ service.nameCap }}_AddService( void )
{
  uint8_t status;

{% for char in service.chars -%}
{% if char.hasCCCD -%}
  // Allocate Client Characteristic Configuration table
  {{ service.nameAbbr}}_{{ char.nameCap }}Config = (gattCharCfg_t *)ICall_malloc( sizeof(gattCharCfg_t) * linkDBNumConns );
  if ( {{ service.nameAbbr}}_{{ char.nameCap }}Config == NULL )
  {     
    return ( bleMemAllocError );
  }
  
  // Initialize Client Characteristic Configuration attributes
  GATTServApp_InitCharCfg( INVALID_CONNHANDLE, {{ service.nameAbbr}}_{{ char.nameCap }}Config );
{% endif -%}
{% endfor -%}
  // Register GATT attribute list and CBs with GATT Server App
  status = GATTServApp_RegisterService( {{service.name}}AttrTbl, 
                                        GATT_NUM_ATTRS( {{service.name}}AttrTbl ),
                                        GATT_MAX_ENCRYPT_KEY_SIZE,
                                        &{{service.name}}CBs );

  return ( status );
}

{% if fluff -%}/*
 * {{ service.nameCap }}_RegisterAppCBs - Registers the application callback function.
 *                    Only call this function once.
 *
 *    appCallbacks - pointer to application callbacks.
 */
{% endif -%}bStatus_t {{ service.nameCap }}_RegisterAppCBs( {{service.nameCap}}CBs_t *appCallbacks )
{
  if ( appCallbacks )
  {
    pAppCBs = appCallbacks;
    
    return ( SUCCESS );
  }
  else
  {
    return ( bleAlreadyInRequestedMode );
  }
}

{% if fluff -%}/*
 * {{ service.nameCap }}_SetParameter - Set a {{ service.nameCap }} parameter.
 *
 *    param - Profile parameter ID
 *    len   - length of data to write
 *    value - pointer to data to write.  This is dependent on
 *            the parameter ID and may be cast to the appropriate 
 *            data type (example: data type of uint16_t will be cast to 
 *            uint16_t pointer).
 */
{% endif -%}
bStatus_t {{ service.nameCap }}_SetParameter( uint8_t param, uint16_t len, void *value )
{
  bStatus_t ret = SUCCESS;
  switch ( param )
  {
{% for char in service.charsNeedRead -%}
    case {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_ID:
      if ( len <= {{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_LEN &&
           len >= {{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_LEN_MIN ) 
      {
        memcpy({{ service.nameAbbr}}_{{ char.nameCap }}Val, value, len);
        {{ service.nameAbbr}}_{{ char.nameCap }}ValLen = len; // Update length for read and get.
{% if char.hasCCCD -%}

        // Try to send notification.
        GATTServApp_ProcessCharCfg( {{ service.nameAbbr}}_{{ char.nameCap }}Config, (uint8_t *)&{{ service.nameAbbr}}_{{ char.nameCap }}Val, FALSE,
                                    {{ service.name}}AttrTbl, GATT_NUM_ATTRS( {{ service.name}}AttrTbl ),
                                    INVALID_TASK_ID,  {{service.name}}_ReadAttrCB);
{% endif -%}
      }
      else
      {
        ret = bleInvalidRange;
      }
      break;

{% endfor -%}
    default:
      ret = INVALIDPARAMETER;
      break;
  }
  return ret;
}


{% if fluff -%}/*
 * {{ service.nameCap }}_GetParameter - Get a {{ service.nameCap }} parameter.
 *
 *    param - Profile parameter ID
 *    len   - pointer to a variable that contains the maximum length that can be written to *value.
              After the call, this value will contain the actual returned length.
 *    value - pointer to data to write.  This is dependent on
 *            the parameter ID and may be cast to the appropriate 
 *            data type (example: data type of uint16_t will be cast to 
 *            uint16_t pointer).
 */
{% endif -%}bStatus_t {{ service.nameCap }}_GetParameter( uint8_t param, uint16_t *len, void *value )
{
  bStatus_t ret = SUCCESS;
  switch ( param )
  {
{% for char in service.charsNeedWrite -%}
    case {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_ID:
      *len = MIN(*len, {{ service.nameAbbr}}_{{ char.nameCap }}ValLen);
      memcpy(value, {{ service.nameAbbr}}_{{ char.nameCap }}Val, *len);
      break;

{% endfor -%}
    default:
      ret = INVALIDPARAMETER;
      break;
  }
  return ret;
}


{% if fluff -%}/*********************************************************************
 * @fn          {{service.name}}_ReadAttrCB
 *
 * @brief       Read an attribute.
 *
 * @param       connHandle - connection message was received on
 * @param       pAttr - pointer to attribute
 * @param       pValue - pointer to data to be read
 * @param       pLen - length of data to be read
 * @param       offset - offset of the first octet to be read
 * @param       maxLen - maximum length of data to be read
 * @param       method - type of read message
 *
 * @return      SUCCESS, blePending or Failure
 */
{% endif -%}static bStatus_t {{service.name}}_ReadAttrCB( uint16_t connHandle, gattAttribute_t *pAttr, 
                                       uint8_t *pValue, uint16_t *pLen, uint16_t offset,
                                       uint16_t maxLen, uint8_t method )
{
  bStatus_t status = SUCCESS;

{% for char in service.charsNeedRead -%}
  // See if request is regarding the {{ char.nameCap }} Characteristic Value
{% if forloop.index0 -%}else {% endif -%}if ( ! memcmp(pAttr->type.uuid, {{ service.nameAbbr}}_{{ char.nameCap }}UUID, pAttr->type.len) )
  {
    if ( offset > {{ service.nameAbbr}}_{{ char.nameCap }}ValLen )  // Prevent malicious ATT ReadBlob offsets.
    {
      status = ATT_ERR_INVALID_OFFSET;
    }
    else
    {
      *pLen = MIN(maxLen, {{ service.nameAbbr}}_{{ char.nameCap }}ValLen - offset);  // Transmit as much as possible
      memcpy(pValue, pAttr->pValue + offset, *pLen);
    }
  }
{% endfor -%}
{% unless service.charsNeedRead == empty-%}
  else
{% endunless -%}
  {
    // If we get here, that means you've forgotten to add an if clause for a
    // characteristic value attribute in the attribute table that has READ permissions.
    *pLen = 0;
    status = ATT_ERR_ATTR_NOT_FOUND;
  }

  return status;
}


{% if fluff -%}/*********************************************************************
 * @fn      {{service.name}}_WriteAttrCB
 *
 * @brief   Validate attribute data prior to a write operation
 *
 * @param   connHandle - connection message was received on
 * @param   pAttr - pointer to attribute
 * @param   pValue - pointer to data to be written
 * @param   len - length of data
 * @param   offset - offset of the first octet to be written
 * @param   method - type of write message
 *
 * @return  SUCCESS, blePending or Failure
 */
{% endif -%}static bStatus_t {{service.name}}_WriteAttrCB( uint16_t connHandle, gattAttribute_t *pAttr,
                                        uint8_t *pValue, uint16_t len, uint16_t offset,
                                        uint8_t method )
{
  bStatus_t status  = SUCCESS;
  uint8_t   paramID = 0xFF;

  // See if request is regarding a Client Characterisic Configuration
  if ( ! memcmp(pAttr->type.uuid, clientCharCfgUUID, pAttr->type.len) )
  {
    // Allow notification and indication, but do not check if really allowed per CCCD.
    status = GATTServApp_ProcessCCCWriteReq( connHandle, pAttr, pValue, len,
                                             offset, GATT_CLIENT_CFG_NOTIFY |
                                                     GATT_CLIENT_CFG_INDICATE );
  }
{% for char in service.charsNeedWrite -%}
  // See if request is regarding the {{ char.nameCap }} Characteristic Value
  else if ( ! memcmp(pAttr->type.uuid, {{ service.nameAbbr}}_{{ char.nameCap }}UUID, pAttr->type.len) )
  {
    if ( offset >= {{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_LEN )
    {
      status = ATT_ERR_INVALID_OFFSET;
    }
    else if ( offset + len > {{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_LEN )
    {
      status = ATT_ERR_INVALID_VALUE_SIZE;
    }
    else if ( offset + len < {{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_LEN_MIN && 
              ( method == ATT_EXECUTE_WRITE_REQ || method == ATT_WRITE_REQ ) )
    {
      // Refuse writes that are lower than minimum.
      // Note: Cannot determine if a Reliable Write (to several chars) is finished, so those will
      //       only be refused if this attribute is the last in the queue (method is execute).
      //       Otherwise, reliable writes are accepted and parsed piecemeal.
      status = ATT_ERR_INVALID_VALUE_SIZE;
    }
    else
    {
      // Copy pValue into the variable we point to from the attribute table.
      memcpy(pAttr->pValue + offset, pValue, len);

      // Only notify application and update length if enough data is written.
      //
      // Note: If reliable writes are used (meaning several attributes are written to using ATT PrepareWrite),
      //       the application will get a callback for every write with an offset + len larger than _LEN_MIN.
      // Note: For Long Writes (ATT Prepare + Execute towards only one attribute) only one callback will be issued,
      //       because the write fragments are concatenated before being sent here.
      if ( offset + len >= {{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_LEN_MIN)
      {
        paramID = {{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_ID;
        {{ service.nameAbbr}}_{{ char.nameCap }}ValLen = offset + len; // Update data length.
      }
    }
  }
{% endfor -%}
  else
  {
    // If we get here, that means you've forgotten to add an if clause for a
    // characteristic value attribute in the attribute table that has WRITE permissions.
    status = ATT_ERR_ATTR_NOT_FOUND;
  }

  // Let the application know something changed (if it did) by using the
  // callback it registered earlier (if it did).
  if (paramID != 0xFF)
    if ( pAppCBs && pAppCBs->pfnChangeCb )
      pAppCBs->pfnChangeCb( paramID ); // Call app function from stack task context.

  return status;
}
